@@ -1,5 +1,5 @@
 /* $Copyright: $
- * Copyright (c) 1996 - 2024 by Steve Baker (steve.baker.llc@gmail.com)
+ * Copyright (c) 1996 - 2026 by Steve Baker (steve.baker.llc@gmail.com)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -17,21 +17,16 @@
  */
 #include "tree.h"
 
-extern char *version, *hversion;
-extern bool dflag, lflag, pflag, sflag, Fflag, aflag, fflag, uflag, gflag;
-extern bool Dflag, inodeflag, devflag, Rflag, duflag, hflag, siflag;
-extern bool noindent, force_color, xdev, nolinks, metafirst, noreport;
+extern struct Flags flag;
+extern char *hversion;
 extern char *host, *sp, *title, *Hintro, *Houtro;
 extern const char *charset;
 
 extern FILE *outfile;
-extern int Level, *dirs, maxdirs;
 
-extern bool colorize, linktargetcolor;
-extern char *endcode;
 extern const struct linedraw *linedraw;
 
-int htmldirlen = 0;
+size_t htmldirlen = 0;
 
 char *class(struct _info *info)
 {
@@ -65,39 +60,25 @@ void html_encode(FILE *fd, char *s)
   }
 }
 
-void url_encode(FILE *fd, char *s)
+// Returns true if the last character printed was a slash
+bool url_encode(FILE *fd, char *s)
 {
+  // Removes / from the reserved list:
+  static const char *reserved = "!#$&'()*+,:;=?@[]";
+  bool slash = false;
+
   for(;*s;s++) {
-    switch(*s) {
-      case ' ':
-      case '"':
-      case '#':
-      case '%':
-      case '<':
-      case '>':
-      case '[':
-      case ']':
-      case '^':
-      case '\\':
-      case '?':
-      case '+':
-	fprintf(fd,"%%%02X",*s);
-	break;
-      case '&':
-	fprintf(fd,"&amp;");
-	break;
-      default:
-	fprintf(fd,isprint((u_int)*s)?"%c":"%%%02X",(u_char)*s);
-	break;
-    }
+    fprintf(fd, (isprint((u_int)*s) && (strchr(reserved, *s) == NULL))? "%c":"%%%02X", *s);
+    slash = (*s == '/');
   }
+  return slash;
 }
 
-void fcat(char *filename)
+void fcat(const char *filename)
 {
   FILE *fp;
   char buf[PATH_MAX];
-  int n;
+  size_t n;
 
   if ((fp = fopen(filename, "r")) == NULL) return;
   while((n = fread(buf, sizeof(char), PATH_MAX, fp)) > 0) {
@@ -117,7 +98,7 @@ void html_intro(void)
 	" <meta http-equiv=\"Content-Type\" content=\"text/html; charset=%s\">\n"
 	" <meta name=\"Author\" content=\"Made by 'tree'\">\n"
 	" <meta name=\"GENERATOR\" content=\"", charset ? charset : "iso-8859-1");
-    print_version(FALSE);
+    print_version(false);
     fprintf(outfile, "\">\n"
 	" <title>%s</title>\n"
 	" <style type=\"text/css\">\n"
@@ -168,17 +149,19 @@ void html_print(char *s)
 
 int html_printinfo(char *dirname, struct _info *file, int level)
 {
+  UNUSED(dirname);
+
   char info[512];
 
   fillinfo(info,file);
-  if (metafirst) {
+  if (flag.metafirst) {
     if (info[0] == '[') {
       html_print(info);
       fprintf(outfile,"%s%s", sp, sp);
     }
-    if (!noindent) indent(level);
+    if (!flag.noindent) indent(level);
   } else {
-    if (!noindent) indent(level);
+    if (!flag.noindent) indent(level);
     if (info[0] == '[') {
       html_print(info);
       fprintf(outfile,"%s%s", sp, sp);
@@ -195,7 +178,7 @@ int html_printfile(char *dirname, char *filename, struct _info *file, int descen
   /* Switch to using 'a' elements only. Omit href attribute if not a link */
   fprintf(outfile,"<a");
   if (file) {
-    if (force_color) fprintf(outfile," class=\"%s\"", class(file));
+    if (flag.force_color) fprintf(outfile," class=\"%s\"", class(file));
     if (file->comment) {
       fprintf(outfile," title=\"");
       for(i=0; file->comment[i]; i++) {
@@ -205,11 +188,12 @@ int html_printfile(char *dirname, char *filename, struct _info *file, int descen
       fprintf(outfile, "\"");
     }
 
-    if (!nolinks) {
+    if (!flag.nolinks) {
       fprintf(outfile," href=\"%s",host);
       if (dirname != NULL) {
-	if (host[strlen(host)-1] != '/') putc('/', outfile);
-	url_encode(outfile, dirname);
+	size_t len = strlen(dirname);
+	size_t off = (len >= htmldirlen? htmldirlen : 0);
+	url_encode(outfile, dirname + (flag.htmloffset? off : 0));
 	if (strcmp(dirname, filename) != 0) {
 	  if (dirname[strlen(dirname)-1] != '/') putc('/', outfile);
 	  url_encode(outfile, filename);
@@ -239,11 +223,15 @@ int html_error(char *error)
 
 void html_newline(struct _info *file, int level, int postdir, int needcomma)
 {
+  UNUSED(file);UNUSED(level);UNUSED(postdir);UNUSED(needcomma);
+
   fprintf(outfile, "<br>\n");
 }
 
 void html_close(struct _info *file, int level, int needcomma)
 {
+  UNUSED(level);UNUSED(needcomma);
+
   fprintf(outfile, "</%s><br>\n", file->tag);
 }
 
@@ -253,11 +241,11 @@ void html_report(struct totals tot)
 
   fprintf(outfile,"<br><br><p>\n\n");
 
-  if (duflag) {
+  if (flag.du) {
     psize(buf, tot.size);
-    fprintf(outfile,"%s%s used in ", buf, hflag || siflag? "" : " bytes");
+    fprintf(outfile,"%s%s used in ", buf, flag.h || flag.si? "" : " bytes");
   }
-  if (dflag)
+  if (flag.d)
     fprintf(outfile,"%ld director%s\n",tot.dirs,(tot.dirs==1? "y":"ies"));
   else
     fprintf(outfile,"%ld director%s, %ld file%s\n",tot.dirs,(tot.dirs==1? "y":"ies"),tot.files,(tot.files==1? "":"s"));
