@@ -1,5 +1,5 @@
 /* $Copyright: $
- * Copyright (c) 1996 - 2024 by Steve Baker (steve.baker.llc@gmail.com)
+ * Copyright (c) 1996 - 2026 by Steve Baker (steve.baker.llc@gmail.com)
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -18,14 +18,13 @@
 #include "tree.h"
 
 extern FILE *outfile;
-extern bool dflag, Fflag, duflag, metafirst, hflag, siflag, noindent;
-extern bool colorize, linktargetcolor, hyperflag;
+extern struct Flags flag;
 extern const struct linedraw *linedraw;
 extern int *dirs;
 extern char *scheme, *authority;
 static char info[512] = {0};
 
-extern char realbasepath[PATH_MAX], xpattern[PATH_MAX];
+extern char realbasepath[PATH_MAX];
 extern size_t dirpathoffset;
 
 int unix_printinfo(char *dirname, struct _info *file, int level)
@@ -33,11 +32,11 @@ int unix_printinfo(char *dirname, struct _info *file, int level)
   UNUSED(dirname);
 
   fillinfo(info, file);
-  if (metafirst) {
+  if (flag.metafirst) {
     if (info[0] == '[') fprintf(outfile, "%s  ",info);
-    if (!noindent) indent(level);
+    if (!flag.noindent) indent(level);
   } else {
-    if (!noindent) indent(level);
+    if (!flag.noindent) indent(level);
     if (info[0] == '[') fprintf(outfile, "%s  ",info);
   }
   return 0;
@@ -47,9 +46,14 @@ void open_hyperlink(char *dirname, char *filename)
 {
   fprintf(outfile,"\033]8;;%s", scheme);
   url_encode(outfile, authority);
-  url_encode(outfile, realbasepath);
-  url_encode(outfile, dirname+dirpathoffset);
-  fputc('/',outfile);
+  fprintf(outfile, ":");
+  // (optional) Hanging slashes are a real pain to deal with
+  bool slash = url_encode(outfile, realbasepath);
+  if (*(dirname+dirpathoffset)) {
+    slash = slash || (*(dirname+dirpathoffset) == '/');
+    if (!slash) fputc('/', outfile);
+    if (!url_encode(outfile, dirname+dirpathoffset)) fputc('/', outfile);
+  } else if (!slash) fputc('/', outfile);
   url_encode(outfile, filename);
   fprintf(outfile,"\033\\");
 }
@@ -66,32 +70,33 @@ int unix_printfile(char *dirname, char *filename, struct _info *file, int descen
   bool colored = false;
   int c;
 
-  if (hyperflag) open_hyperlink(dirname, file->name);
+  if (file) {
+    if (flag.hyper) open_hyperlink(dirname, file->name);
 
-  if (file && colorize) {
-    if (file->lnk && linktargetcolor) colored = color(file->lnkmode, file->name, file->orphan, false);
-    else colored = color(file->mode, file->name, file->orphan, false);
+    if (flag.colorize) {
+      if (file->lnk && flag.linktargetcolor) colored = color(file->lnkmode, file->name, file->orphan, false);
+      else colored = color(file->mode, file->name, file->orphan, false);
+    }
   }
 
   printit(filename);
-
   if (colored) endcolor();
 
-  if (hyperflag) close_hyperlink();
-
   if (file) {
-    if (Fflag && !file->lnk) {
+    if (flag.hyper) close_hyperlink();
+
+    if (flag.F && !file->lnk) {
       if ((c = Ftype(file->mode))) fputc(c, outfile);
     }
 
     if (file->lnk) {
       fprintf(outfile," -> ");
-      if (hyperflag) open_hyperlink(dirname, file->name);
-      if (colorize) colored = color(file->lnkmode, file->lnk, file->orphan, true);
+      if (flag.hyper) open_hyperlink(dirname, file->name);
+      if (flag.colorize) colored = color(file->lnkmode, file->lnk, file->orphan, true);
       printit(file->lnk);
       if (colored) endcolor();
-      if (hyperflag) close_hyperlink();
-      if (Fflag) {
+      if (flag.hyper) close_hyperlink();
+      if (flag.F) {
 	if ((c = Ftype(file->lnkmode))) fputc(c, outfile);
       }
     }
@@ -112,12 +117,12 @@ void unix_newline(struct _info *file, int level, int postdir, int needcomma)
   if (postdir <= 0) fprintf(outfile, "\n");
   if (file && file->comment) {
     size_t infosize = 0, line, lines;
-    if (metafirst) infosize = info[0] == '['? strlen(info)+2 : 0;
+    if (flag.metafirst) infosize = info[0] == '['? strlen(info)+2 : 0;
 
     for(lines = 0; file->comment[lines]; lines++);
     dirs[level+1] = 1;
     for(line = 0; line < lines; line++) {
-      if (metafirst) {
+      if (flag.metafirst) {
 	printf("%*s", (int)infosize, "");
       }
       indent(level);
@@ -132,11 +137,11 @@ void unix_report(struct totals tot)
   char buf[256];
 
   fputc('\n', outfile);
-  if (duflag) {
+  if (flag.du) {
     psize(buf, tot.size);
-    fprintf(outfile,"%s%s used in ", buf, hflag || siflag? "" : " bytes");
+    fprintf(outfile,"%s%s used in ", buf, flag.h || flag.si? "" : " bytes");
   }
-  if (dflag)
+  if (flag.d)
     fprintf(outfile,"%ld director%s\n",tot.dirs,(tot.dirs==1? "y":"ies"));
   else
     fprintf(outfile,"%ld director%s, %ld file%s\n",tot.dirs,(tot.dirs==1? "y":"ies"),tot.files,(tot.files==1? "":"s"));
